from datetime import datetime
import random
import re


def _generate_slug(length=10):
    letters = "2346789abcdefghijkmnpqrtwxyz"
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str


def safe_id_str(s):
    s = s.lower()
    return re.sub(r'[^a-z0-9_]', "_", s)


def create_safe_id(docType, uid=None):
    return f"{docType}:{safe_id_str(uid)}" if uid else generate_random_id(docType)


def generate_random_id(docType):
    return f"{docType}:{int(datetime.utcnow().timestamp())}_{_generate_slug()}"


def create_doc(docType, data):
    return {
        '_id': data.get("id", generate_random_id(docType)),
        'type': docType,
        'created_at': datetime.now().astimezone().replace(microsecond=0).isoformat(),
        "schema": "0",
        docType: data
    }


def create_tree(name, nodes, nodeDocType=""):
    return {
        'id': f"tree:{name}",
        'nodes': nodes,
        'nodeDocType': nodeDocType
    }


def create_tree_node(docType, label, parent=None, child_nodes=[]):
    return {
        'id': generate_random_id(docType),
        'label': label,
        'parentId': parent,
        'childNodes': child_nodes
    }



import argparse
from .legacy import dump1


def cli():
    parser = argparse.ArgumentParser(description="BOMIST Legacy Utilities")
    parser.add_argument("--dump1", default=False, action="store_true")
    parser.add_argument("--ws", help="Workspace path")
    parser.add_argument("--out", help="Output file path")
    args = parser.parse_args()

    if args.dump1:
        if not args.ws:
            parser.error("--ws is required")
        dump1(args.ws, args.out)
    else:
        parser.print_help()


if __name__ == "__main__":
    cli()



"""
BOMIST Utilities (bomist.com)
"""

import json
import shutil
import tarfile
import re
from glob import glob
from os import path, getcwd, listdir, makedirs

from .crypto import decode
from .v2 import (create_doc, create_tree_node, create_tree,
                 create_safe_id, generate_random_id, safe_id_str)


def dump1(wspath, outpath=None):
    print(f"Dumping {wspath}")
    if not _check_if_ws_exits(wspath):
        print("ERROR", f"workspace not found at {wspath}",)
        return

    datapath = path.join(wspath, "data")

    all_docs = []

    storage_mapper = {}
    labels_mapper = {}
    category_mapper = {}
    docs_mapper = {}  # ref -> id
    files_mapper = {}  # file_main -> id

    with open(path.join(datapath, ".labels"), "rb") as f:
        d = json.loads(decode(f.read()))
        plabels = d.get("labels", [])
        nodes = []
        for plabel in plabels:
            pname = plabel.get("name")
            pnode = create_tree_node("label", pname)
            k = pname
            labels_mapper[k] = pnode['id']
            child_nodes = []
            for slabel in plabel.get("sub", []):
                sname = slabel.get('name')
                snode = create_tree_node(
                    "label", sname, pnode['id'])
                child_nodes.append(snode)
                k = "\t".join([pname, sname])
                labels_mapper[k] = snode['id']
            pnode['childNodes'] = child_nodes
            nodes.append(pnode)

        labels_tree_doc = create_doc("tree", create_tree("labels", nodes))
        all_docs += [labels_tree_doc]

    with open(path.join(datapath, ".storage"), "rb") as f:
        d = json.loads(decode(f.read()))
        [tree_doc, storage_docs, mapper] = _create_storage(d)
        storage_mapper = mapper
        all_docs += [tree_doc] + storage_docs

    with open(path.join(datapath, ".categories"), "rb") as f:
        category_docs = []
        d = json.loads(decode(f.read()))
        categories = d["categories"]
        for c in categories:
            doc = create_doc("category", {
                'name': c
            })
            category_mapper[c] = doc['_id']
            category_docs.append(doc)
        all_docs += category_docs

    docpaths = listdir(path.join(datapath, "files"))
    for docpath in docpaths:
        doc_docs = []
        with open(path.join(datapath, "files", docpath, ".doc"), "rb") as f:
            d = json.loads(decode(f.read()))
            doc = create_doc("document", {})
            ext = path.splitext(d['file_main'])[1]
            fpath = f"{safe_id_str(doc['_id'])}{ext}"
            doc.update({
                'document': {
                    'name': d['title'],
                    'notes': d['notes'],
                    'category': category_mapper.get(d['category'], "") if d['category'] else "",
                    'path': {
                        'local': fpath
                    }
                }
            })
            k = d['ref']
            docs_mapper[k] = doc['_id']
            k = d['file_main']
            files_mapper[k] = fpath
            doc_docs.append(doc)
        all_docs += doc_docs

    partpaths = listdir(path.join(datapath, "parts"))
    for partpath in partpaths:
        part_docs = []
        with open(path.join(datapath, "parts", partpath, ".part"), "rb") as f:
            d = json.loads(decode(f.read()))

            label_k = "\t".join([d['plabel'], d['slabel']])
            label = labels_mapper.get(label_k, "") if label_k else ""

            storage_k = d['storage']
            storage = storage_mapper.get(storage_k) if storage_k else ""

            re_tolerance = re.compile(
                "([0-9.,]+)\s*(?=%|ppm)").findall(d['value'] or "")
            tolerance = int(re_tolerance[0]) if len(re_tolerance) > 0 else None

            part = {
                'id': create_safe_id("part", f"{d['mpn']}_{d['brand']}"),
                'type': "outsourced",
                'mpn': d['mpn'],
                'manufacturer': d['brand'],
                'description': d['desc'],
                'value': d['value'],
                'package': d['package'],
                'tolerance': tolerance,
                'label': label,
                'stock': int(d['stock']),
                'stockInHouse': int(d['stock']),
                'lowStock': int(d['stock_low']),
                'notes': d['notes'],
                'hidden': bool(d.get('obsolete', False) or d['stock_ignore'])
            }
            doc = create_doc("part", part)

            inventory_doc = create_doc("inventory", {
                'id': f"r/{part['id']}:{generate_random_id('inventory')}",
                'status': 'in_house',
                'storage': storage,
                'qty': int(d['stock'])
            })

            if d['docs']:
                for ref in d['docs']:
                    doc_id = docs_mapper.get(ref)
                    ref_doc = create_doc("ref", {
                        'id': f"r/{part['id']}:{doc_id}",
                        'refId': doc_id
                    })
                    part_docs += [ref_doc]

            part_docs += [doc, inventory_doc]

        all_docs += part_docs

    # projectpaths = listdir(path.join(datapath, "projects"))
    # for projectpath in projectpaths:
    #     project_data = {}
    #     projectpath = path.join(datapath, "projects", projectpath)
    #     with open(path.join(projectpath, ".project"), "rb") as f:
    #         d = json.loads(decode(f.read()))
    #         project_data = d
    #     print("--- project", project_data)

    #     revpaths = glob(path.join(projectpath, "rev_*"))
    #     # print("revs", revpaths)
    #     for revpath in revpaths:
    #         rev_data = {}
    #         with open(path.join(revpath, ".rev"), "rb") as f:
    #             d = json.loads(decode(f.read()))
    #             rev_data = d
    #         print("rev", rev_data)
    #         varpaths = glob(path.join(projectpath, revpath, "var_*"))
    #         print("varpaths", varpaths)
    #         for varpath in varpaths:
    #             var_data = {}
    #             with open(path.join(varpath, ".var"), "rb") as f:
    #                 d = json.loads(decode(f.read()))
    #                 var_data = d
    #             print("var", json.dumps(var_data, indent=2))
                

    outpath = outpath or getcwd()

    tmp_dir = path.join(outpath, ".tmp")
    dump_dir = path.join(tmp_dir, "dump-data")
    files_dir = path.join(dump_dir, "files")

    makedirs(files_dir, exist_ok=True)
    makedirs(outpath, exist_ok=True)

    with open(path.join(dump_dir, "allDocs.json"), "w") as f:
        rows = list(map(lambda d: dict({
            'doc': d
        }), all_docs))
        all_docs = {
            'total_rows': len(all_docs),
            'rows': rows
        }
        f.write(json.dumps(all_docs, indent=2))

    docpaths = listdir(path.join(datapath, "files"))
    for docpath in docpaths:
        filepaths = listdir(path.join(datapath, "files", docpath))
        for filepath in filepaths:
            if filepath in files_mapper.keys():
                fsrc = path.join(datapath, "files", docpath, filepath)
                fdst = path.join(files_dir, files_mapper.get(filepath))
                shutil.copy2(fsrc, fdst)

    tarfilename = "legacy.bomist_dump"
    tarpath = path.join(outpath, tarfilename)
    with tarfile.open(tarpath, "w:gz") as tar:
        tar.add(dump_dir, arcname="dump-data")

    shutil.rmtree(tmp_dir)

    print(f"{tarfilename} created")


def _create_storage(storage):
    mapper = {}
    storage_docs = []

    def _for_each_storage(storage, parent=None, parent_path=None):
        nodes = []
        for s in storage:
            values = s.get('data', [])
            name = values[0]
            spath = (parent_path or []) + [name]
            node = create_tree_node("storage", name, parent)
            k = "\t".join(spath)
            mapper[k] = node['id']
            sdoc = create_doc("storage", {'name': name})
            storage_docs.append(sdoc)
            node['childNodes'] = _for_each_storage(
                s.get('children', []), node['id'], spath)
            nodes.append(node)
        return nodes

    tree_nodes = _for_each_storage(storage, None)
    storage_tree_doc = create_doc("tree", create_tree("storage", tree_nodes))
    return [storage_tree_doc, storage_docs, mapper]


def _check_if_ws_exits(wspath):
    wsfile = ".ws"
    wspath = path.join(wspath, wsfile)
    return path.isfile(wspath)



import argparse
from random import randint

BOMER_CRYPTO_MASKS_COUNT = 30
BOMER_CRYPTO_MAGIC = 0xA5
BOMER_CRYPTO_MASKS = [
    0x4F, 0xAA, 0xC4, 0xF3, 0x45,
    0x23, 0xCD, 0xBE, 0xEF, 0xCC,
    0xF2, 0xBB, 0xDE, 0xAD, 0x9E,
    0xC3, 0xAA, 0xC4, 0xF3, 0x5D,
    0x12, 0x3B, 0xAE, 0xBE, 0xEF,
    0x23, 0xCD, 0xBE, 0xEF, 0xCC
]


def encode(data):
    data = bytearray(data, encoding="utf-8")
    checksum = 0
    encoded = []
    data_len = len(data)
    mask_index = randint(0, BOMER_CRYPTO_MASKS_COUNT - 1)
    encoded.append((mask_index ^ BOMER_CRYPTO_MAGIC))
    for b in data:
        b = int(b)
        mask = BOMER_CRYPTO_MASKS[mask_index]
        encoded_byte = (b & 0xFF) ^ mask
        encoded.append(encoded_byte & 0xFF)
        checksum += encoded_byte
        mask_index = (mask_index + 1) % BOMER_CRYPTO_MASKS_COUNT
    encoded.append(((checksum & 0xFF) ^ (0xFF - BOMER_CRYPTO_MAGIC)) & 0xFF)
    return bytearray(encoded)


def decode(data):
    data = bytearray(data)
    checksum = 0
    encoded = []
    data_len = len(data)
    mask_index = (data[0] ^ BOMER_CRYPTO_MAGIC) & 0xFF
    for b in data[1:data_len - 1]:
        mask = BOMER_CRYPTO_MASKS[mask_index]
        checksum += b & 0xFF
        encoded_byte = (b ^ mask) & 0xFF
        encoded.append(encoded_byte)
        mask_index = (mask_index + 1) % BOMER_CRYPTO_MASKS_COUNT
    checksum = (checksum & 0xFF) ^ (0xFF - BOMER_CRYPTO_MAGIC)
    data_checksum = data[-1]
    if checksum != data_checksum:
        print("!!")
    return bytearray(encoded).decode("utf-8")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bomist_utils']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['bomist_utils = bomist_utils:cli']}

setup_kwargs = {
    'name': 'bomist-utils',
    'version': '0.1.8',
    'description': 'BOMIST Utilities',
    'long_description': "# BOMIST Utilities\n\nThings you can do with it:\n\n- Convert legacy workspaces into BOMIST v2 workspaces\n\n## How to install\n\nThis requires Python 3 and the Python Package Installer (pip3) to be installed.\nYou can download it here:\nhttps://www.python.org/downloads/\n\nOnce you have them installed, run the following command on your terminal:\n\n```\n$ pip3 install bomist_utils\n```\n\nAfter installing `bomist_utils` through `pip3` you'll end up with it available on your terminal.\n\nOn Windows you might have to re-launch your terminal in order for the `bomist_utils` command to be recognized.\n\nTo make sure it was properly installed run the following command:\n\n```\n$ bomist_utils --help\n```\n\nYou should see the available options.\n\n## Usage\n\n### Convert legacy workspaces\n\n```\n$ bomist_utils --dump1 --ws <wspath> [--out <outpath>]\n```\n\n`wspath` is the path of the workspace you want to dump. A `.ws` file must exist in it.\n\nA `legacy.bomist_dump` file will be created on `outpath` or in the folder the command was called from if the `--out` option is not used. This file can then be imported by BOMIST v2.\n\n#### **Limitations**\n\nAt the moment this utility can only convert and keep data connections between:\n\n```\nparts, documents, labels, storage, categories\n```\n\nProjects, orders and history won't be converted.\n\nTo import projects into the new version you can export your BOMs into CSV files in the legacy version (right-click > Export) and then import them in the new version. Parts will be accordingly assigned to BOM designators, as long as those parts exists in the workspace.\n\nNeed help? [Get in touch](https://bomist.com/support/contact/)\n\n---\n\nFor more info: [bomist.com](https://bomist.com)\n",
    'author': 'Mario Ribeiro',
    'author_email': 'mario@bomist.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://bomist.com',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
